from slicerator import pipeline
from pims.base_frames import FramesSequence, FramesSequenceND
from pims.frame import Frame
from pims.display import (export, play, scrollable_stack, to_rgb, normalize,
                          plot_to_frame, plots_to_frame)
from itertools import chain

import glob
import os
from warnings import warn

# has to be here for API stuff
from pims.image_sequence import ImageSequence, ImageSequenceND, ReaderSequence  # noqa
from pims.image_reader import ImageReader, ImageReaderND  # noqa
from .cine import Cine  # noqa
from .norpix_reader import NorpixSeq  # noqa
from pims.tiff_stack import TiffStack_tifffile  # noqa
from .spe_stack import SpeStack
from pims.process import as_grey, as_gray


def not_available(requirement):
    def raiser(*args, **kwargs):
        raise ImportError(
            "This reader requires {0}.".format(requirement))
    return raiser

if export is None:
    export = not_available("PyAV or MoviePy")

try:
    import pims.pyav_reader
    if pims.pyav_reader.available():
        PyAVReaderTimed = pims.pyav_reader.PyAVReaderTimed
        PyAVReaderIndexed = pims.pyav_reader.PyAVReaderIndexed
        Video = PyAVReaderTimed
    else:
        raise ImportError()
except (ImportError, IOError):
    PyAVVideoReader = not_available("PyAV")
    PyAVReaderTimed = not_available("PyAV")
    PyAVReaderIndexed = not_available("PyAV")
    Video = None

PyAVVideoReader = PyAVReaderTimed


try:
    import pims.imageio_reader
    if pims.imageio_reader.available():
        ImageIOReader = pims.imageio_reader.ImageIOReader
        if Video is None:
            if pims.imageio_reader.ffmpeg_available():
                Video = ImageIOReader
    else:
        raise ImportError()
except (ImportError, IOError):
    ImageIOReader = not_available("ImageIO")


try:
    import pims.moviepy_reader
    if pims.moviepy_reader.available():
        MoviePyReader = pims.moviepy_reader.MoviePyReader
        if Video is None:
            Video = MoviePyReader
    else:
        raise ImportError()
except (ImportError, IOError):
    MoviePyReader = not_available("MoviePy")

if Video is None:
    Video = not_available("PyAV, MoviePy, or ImageIO")

import pims.tiff_stack
from pims.tiff_stack import (TiffStack_pil, TiffStack_libtiff,
                                TiffStack_tifffile)
# First, check if each individual class is available
# and drop in placeholders as needed.
if not pims.tiff_stack.tifffile_available():
    TiffStack_tiffile = not_available("tifffile")
if not pims.tiff_stack.libtiff_available():
    TiffStack_libtiff = not_available("libtiff")
if not pims.tiff_stack.PIL_available():
    TiffStack_pil = not_available("PIL or Pillow")
# Second, decide which class to assign to the
# TiffStack alias.
if pims.tiff_stack.tifffile_available():
    TiffStack = TiffStack_tifffile
elif pims.tiff_stack.libtiff_available():
    TiffStack = TiffStack_libtiff
elif pims.tiff_stack.PIL_available():
    TiffStack = TiffStack_pil
else:
    TiffStack = not_available("tifffile, libtiff, or PIL/Pillow")


try:
    import pims.bioformats
    if pims.bioformats.available():
        Bioformats = pims.bioformats.BioformatsReader
    else:
        raise ImportError()
except (ImportError, IOError):
    BioformatsRaw = not_available("JPype")
    Bioformats = not_available("JPype")


try:
    from pims_nd2 import ND2_Reader as ND2Reader_SDK

    class ND2_Reader(ND2Reader_SDK):
        class_priority = 0

        def __init__(self, *args, **kwargs):
            warn("'ND2_Reader' has been renamed to 'ND2Reader_SDK' and will be"
                 "removed in future pims versions. "
                 "Please use the new name, or try out the pure-Python one named "
                 "`ND2Reader`.")
            super(ND2_Reader, self).__init__(*args, **kwargs)
except ImportError:
    ND2Reader_SDK = not_available("pims_nd2")
    ND2_Reader = not_available("pims_nd2")

try:
    from nd2reader import ND2Reader
except ImportError:
    ND2Reader = not_available("nd2reader")

def open(sequence, **kwargs):
    """Read a filename, list of filenames, or directory of image files into an
    iterable that returns images as numpy arrays.

    Parameters
    ----------
    sequence : string, list of strings, or glob
        The sequence you want to load. This can be a directory containing
        images, a glob ('/path/foo*.png') pattern of images,
        a video file, or a tiff stack
    kwargs :
        All keyword arguments will be passed to the reader.

    Examples
    --------
    >>> video = open('path/to/images/*.png')  # or *.tif, or *.jpg
    >>> imshow(video[0]) # Show the first frame.
    >>> imshow(video[-1]) # Show the last frame.
    >>> imshow(video[1][0:10, 0:10]) # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video
    """
    files = glob.glob(sequence)
    if len(files) > 1:
        # todo: test if ImageSequence can read the image type,
        #       delegate to subclasses as needed
        return ImageSequence(sequence, **kwargs)

    _, ext = os.path.splitext(sequence)
    if ext is None or len(ext) < 2:
        raise UnknownFormatError(
            "Could not detect your file type because it did not have an "
            "extension. Try specifying a loader class, e.g. "
            "Video({0})".format(sequence))
    ext = ext.lower()[1:]

    # list all readers derived from the pims baseclasses
    all_handlers = chain(_recursive_subclasses(FramesSequence),
                         _recursive_subclasses(FramesSequenceND))
    # keep handlers that support the file ext. use set to avoid duplicates.
    eligible_handlers = set(h for h in all_handlers
                            if ext and ext in map(_drop_dot, h.class_exts()))
    if len(eligible_handlers) < 1:
        raise UnknownFormatError(
            "Could not autodetect how to load a file of type {0}. "
            "Try manually "
            "specifying a loader class, e.g. Video({1})".format(ext, sequence))

    def sort_on_priority(handlers):
        # This uses optional priority information from subclasses
        # > 10 means that it will be used instead of than built-in subclasses
        def priority(cls):
            try:
                return cls.class_priority
            except AttributeError:
                return 10
        return sorted(handlers, key=priority, reverse=True)

    exceptions = ''
    for handler in sort_on_priority(eligible_handlers):
        try:
            return handler(sequence, **kwargs)
        except Exception as e:
            message = '{0} errored: {1}'.format(str(handler), str(e))
            warn(message)
            exceptions += message + '\n'
    raise UnknownFormatError("All handlers returned exceptions:\n" + exceptions)


class UnknownFormatError(Exception):
    pass


def _recursive_subclasses(cls):
    "Return all subclasses (and their subclasses, etc.)."
    # Source: http://stackoverflow.com/a/3862957/1221924
    return (cls.__subclasses__() +
        [g for s in cls.__subclasses__() for g in _recursive_subclasses(s)])

def _drop_dot(s):
    if s.startswith('.'):
        return s[1:]
    else:
        return s



# The MIT License (MIT)
#
# Copyright (c) 2014 Zulko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# The MIT License (MIT)
# [OSI Approved License]
#
# https://github.com/Zulko/moviepy
#
# source files:
# moviepy/conf.py
# moviepy/video/io/ffmpeg_reader.py
# moviepy/tools.py
#
# Files heavily edited by PIMS contributors
# January 2014

import re
import subprocess as sp
import sys
import os

import numpy as np

from pims.base_frames import FramesSequence
from pims.frame import Frame


try:
    from subprocess import DEVNULL  # py3k
except ImportError:
    DEVNULL = open(os.devnull, 'wb')


def try_ffmpeg(FFMPEG_BINARY):
    try:
        proc = sp.Popen([FFMPEG_BINARY],
                        stdout=sp.PIPE,
                        stderr=sp.PIPE)
        proc.wait()
    except:
        return False
    else:
        return True


# Name (and location if needed) of the FFMPEG binary. It will be
# "ffmpeg" on linux, certainly "ffmpeg.exe" on windows, else any path.
FFMPEG_BINARY_SUGGESTIONS = ['ffmpeg', 'ffmpeg.exe']

FFMPEG_BINARY = None
for name in FFMPEG_BINARY_SUGGESTIONS:
    if try_ffmpeg(name):
        FFMPEG_BINARY = name
        break


def available():
    return FFMPEG_BINARY is not None

_pix_fmt_dict = {'rgb24': 3,
                 'rgba': 4}


class FFmpegVideoReader(FramesSequence):
    """Read images from the frames of a standard video file into an
    iterable object that returns images as numpy arrays.

    This reader, based on ffmpeg, should be able to read most video
    files

    Parameters
    ----------
    filename : string
    pix_fmt : string, optional
        Expected number of color channels. Either 'rgb24' or 'rgba'.
        Defaults to 'rgba'.
    use_cache : boolean, optional
        Saves time if the file was previously opened. Defaults to True.
        Set to False if the file may have changed since it was
        last read.

    Examples
    --------
    >>> video = FFmpegVideoReader('video.avi')  # or .mov, etc.
    >>> imshow(video[0]) # Show the first frame.
    >>> imshow(video[-1]) # Show the last frame.
    >>> imshow(video[1][0:10, 0:10]) # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video

    """
    def __init__(self, filename, pix_fmt="rgb24", use_cache=True):

        self.filename = filename
        self.pix_fmt = pix_fmt
        self._initialize(use_cache)
        try:
            self.depth = _pix_fmt_dict[pix_fmt]
        except KeyError:
            raise ValueError("invalid pixel format")
        w, h = self._size
        self._stride = self.depth*w*h

    def _initialize(self, use_cache):
        """ Opens the file, creates the pipe. """

        buffer_filename = '{0}.pims_buffer'.format(self.filename)
        meta_filename = '{0}.pims_meta'.format(self.filename)

        cmd = [FFMPEG_BINARY, '-i', self.filename,
                '-f', 'image2pipe',
                "-pix_fmt", self.pix_fmt,
                '-vcodec', 'rawvideo', '-']
        proc = sp.Popen(cmd, stdin=sp.PIPE,
                             stdout=sp.PIPE,
                             stderr=sp.PIPE)

        print("Decoding video file...")

        if (os.path.isfile(buffer_filename) and os.path.isfile(meta_filename)
            and use_cache):
            print("Reusing buffer from previous opening of this video.")
            self.data_buffer = open(buffer_filename, 'rb')
            self.metafile = open(meta_filename, 'r')
            self._len = int(self.metafile.readline())
            w = int(self.metafile.readline())
            h = int(self.metafile.readline())
            self._size = [w, h]
            return

        self.data_buffer = open(buffer_filename, 'wb')
        self.metafile = open(meta_filename, 'w')
        print ("Decoding video file. This is slow, but only the first time.")
        sys.stdout.flush()
        CHUNKSIZE = 2**14  # utterly arbitrary
        while True:
            try:
                chunk = proc.stdout.read(CHUNKSIZE)
                if len(chunk) == 0:
                    break
                self.data_buffer.write(chunk)
            except EOFError:
                break
        self.data_buffer.close()
        self.data_buffer = open(buffer_filename, 'rb')

        self._process_ffmpeg_stderr(proc.stderr.read())

        proc.terminate()
        for std in proc.stdin, proc.stdout, proc.stderr:
            std.close()

    def _process_ffmpeg_stderr(self, stderr, verbose=False):
        if verbose:
            print(stderr)

        lines = stderr.splitlines()
        if "No such file or directory" in lines[-1]:
            raise IOError("%s not found ! Wrong path ?" % self.filename)

        # get the output lines that describe the video
        line = [l for l in lines if ' Video: ' in l][0]
        # logic to parse all of the MD goes here

        # get the size, of the form 460x320 (w x h)
        match = re.search(" [0-9]*x[0-9]*(,| )", line)
        self._size = map(int, line[match.start():match.end()-1].split('x'))
        # this needs to be more robust
        self._len = int(lines[-2].split()[1])
        self.metafile.write('{0}\n'.format(self._len))
        self.metafile.write('{0}\n'.format(self._size[0]))
        self.metafile.write('{0}\n'.format(self._size[1]))
        self.metafile.close()

    def __len__(self):
        return self._len

    @property
    def frame_shape(self):
        return self._size

    def get_frame(self, j):
        self.data_buffer.seek(self._stride*j)
        s = self.data_buffer.read(self._stride)
        w, h = self._size
        result = np.frombuffer(s,
            dtype='uint8').reshape((h, w, self.depth))
        return Frame(result, frame_no=j)

    @property
    def pixel_type(self):
        raise np.uint8

    @classmethod
    def class_exts(cls):
        return set(['mov', 'avi', 'webm'])

    def close(self):
        self.metafile.close()
        self.data_buffer.close()
        super().close()

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {frame_shape!r}
Pixel Format: {pix_fmt}""".format(frame_shape=self.frame_shape,
                                  count=len(self),
                                  filename=self.filename,
                                  pix_fmt=self.pix_fmt)



import os
from datetime import datetime
import itertools
import warnings

import numpy as np

from pims.frame import Frame

try:
    from PIL import Image  # should work with PIL or PILLOW
except ImportError:
    Image = None

try:
    from libtiff import TIFF
except ImportError:
    TIFF = None

try:
    import tifffile
except ImportError:
    tifffile = None


def libtiff_available():
    return TIFF is not None


def PIL_available():
    return Image is not None


def tifffile_available():
    return tifffile is not None


from pims.base_frames import FramesSequence

_dtype_map = {4: np.uint8,
              8: np.uint8,
              16: np.uint16}

def _tiff_datetime(dt_str):
    """Convert the DateTime string of TIFF files to a datetime object"""
    return datetime(year=int(dt_str[0:4]), month=int(dt_str[5:7]),
                    day=int(dt_str[8:10]), hour=int(dt_str[11:13]),
                    minute=int(dt_str[14:16]), second=int(dt_str[17:19]))


class TiffStack_tifffile(FramesSequence):
    """Read TIFF stacks (single files containing many images) into an
    iterable object that returns images as numpy arrays.

    This reader, based on tiffile.py, should read standard TIFF
    files and sundry derivatives of the format used in microscopy.

    Parameters
    ----------
    filename : string

    Examples
    --------
    >>> video = TiffStack('many_images.tif')  # or .tiff
    >>> imshow(video[0]) # Show the first frame.
    >>> imshow(video[-1]) # Show the last frame.
    >>> imshow(video[1][0:10, 0:10]) # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video

    Note
    ----
    This wraps tifffile.py. It should deal with a range of
    tiff files and sundry microscope related tiff derivatives.
    The obvious thing to do here is to extend tifffile.TiffFile;
    however that would over-ride our nice slicing sematics. The
    way that TiffFile deals with __getitem__ is to just pass it
    through to an underlying list.  Further, it return TiffPages,
    not arrays as we desire.

    See Also
    --------
    TiffStack_pil, TiffStack_libtiff, ImageSequence
    """
    @classmethod
    def class_exts(cls):
        # TODO extend this set to match reality
        return {'tif', 'tiff', 'lsm',
                'stk'} | super(TiffStack_tifffile, cls).class_exts()

    def __init__(self, filename):
        self._filename = filename
        record = tifffile.TiffFile(filename).series[0]
        if hasattr(record, 'pages'):
            self._tiff = record.pages
        else:
            self._tiff = record['pages']

        tmp = self._tiff[0]
        self._dtype = tmp.dtype
        self._im_sz = tmp.shape

    def get_frame(self, j):
        t = self._tiff[j]
        data = t.asarray()
        return Frame(data, frame_no=j, metadata=self._read_metadata(t))

    def _read_metadata(self, tiff):
        """Read metadata for current frame and return as dict"""
        # tags are only stored as a TiffTags object on the parent TiffPage now
        tags = tiff.keyframe.tags
        md = {}
        for name in ('ImageDescription', 'image_description'):
            if name in tags:
                md["ImageDescription"] = tags[name].value
                break
        for name in ('DateTime', 'datetime'):
            if name in tags:
                md["DateTime"] = tags[name].value
                break
        for name in ('Software', 'software'):
            if name in tags:
                md["Software"] = tags[name].value
                break
        for name in ('DocumentName', 'document_name'):
            if name in tags:
                md["DocumentName"] = tags[name].value
                break
        for key in md:
            try:
                md[key] = md[key].decode()
            except AttributeError:
                pass
            except UnicodeDecodeError:
                md[key] = ''
            if key == 'DateTime':
                try:
                    md[key] = _tiff_datetime(md[key])
                except ValueError:
                    warnings.warn(
                        "DateTime tiff tag could not be parsed correctly:" + \
                        str(md[key]))
        return md

    @property
    def pixel_type(self):
        return self._dtype

    @property
    def frame_shape(self):
        return self._im_sz

    def __len__(self):
        return len(self._tiff)

    def close(self):
        self._tiff.parent.close()
        super().close()

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {frame_shape!r}
Pixel Datatype: {dtype}""".format(frame_shape=self.frame_shape,
                                  count=len(self),
                                  filename=self._filename,
                                  dtype=self.pixel_type)


class TiffStack_libtiff(FramesSequence):
    """Read TIFF stacks (single files containing many images) into an
    iterable object that returns images as numpy arrays.

    This reader, based on libtiff, should read standard TIFF files.

    Parameters
    ----------
    filename : string

    Examples
    --------
    >>> video = TiffStack('many_images.tif')  # or .tiff
    >>> imshow(video[0]) # Show the first frame.
    >>> imshow(video[-1]) # Show the last frame.
    >>> imshow(video[1][0:10, 0:10]) # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video

    See Also
    --------
    TiffStack_pil, TiffStack_tiffile, ImageSequence
    """
    def __init__(self, filename):
        self._filename = filename
        self._tiff = TIFF.open(filename)

        self._count = 1
        while not self._tiff.LastDirectory():
            self._count += 1
            self._tiff.ReadDirectory()

        # reset to 0
        self._tiff.SetDirectory(0)

        tmp = self._tiff.read_image()
        self._dtype = tmp.dtype

        self._im_sz = tmp.shape

        self._byte_swap = bool(self._tiff.IsByteSwapped())

    def get_frame(self, j):
        if j > self._count:
            raise ValueError("File does not contain this many frames")
        self._tiff.SetDirectory(j)
        res = self._tiff.read_image()
        if self._byte_swap:
            res = res.newbyteorder()
        return Frame(res, frame_no=j, metadata=self._read_metadata())

    def _read_metadata(self):
        """Read metadata for current frame and return as dict"""
        md = {}
        #explicit str() is needed for python2, otherwise the type would be
        #unicode and then it won't work anymore
        dt = self._tiff.GetField(str("ImageDescription"))
        if dt is not None:
            try:
                md["ImageDescription"] = dt.decode()
            except:
                pass
        dt = self._tiff.GetField(str("DateTime"))
        if dt is not None:
            try:
                md["DateTime"] = _tiff_datetime(dt.decode())
            except:
                pass
        dt = self._tiff.GetField(str("Software"))
        if dt is not None:
            try:
                md["Software"] = dt.decode()
            except:
                pass
        dt = self._tiff.GetField(str("DocumentName"))
        if dt is not None:
            try:
                md["DocumentName"] = dt.decode()
            except:
                pass
        return md

    @property
    def pixel_type(self):
        return self._dtype

    @property
    def frame_shape(self):
        return self._im_sz

    def __len__(self):
        return self._count

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {w} x {h}
Pixel Datatype: {dtype}""".format(w=self.frame_shape[0],
                                  h=self.frame_shape[1],
                                  count=len(self),
                                  filename=self._filename,
                                  dtype=self.pixel_type)

    def close(self):
        self._tiff.close()
        super().close()


class TiffStack_pil(FramesSequence):
    """Read TIFF stacks (single files containing many images) into an
    iterable object that returns images as numpy arrays.

    This reader, based on PIL/Pillow, should read standard TIFF stacks.

    Parameters
    ----------
    filename : string

    Examples
    --------
    >>> video = TiffStack('many_images.tif')  # or .tiff
    >>> imshow(video[0]) # Show the first frame.
    >>> imshow(video[-1]) # Show the last frame.
    >>> imshow(video[1][0:10, 0:10]) # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video

    See Also
    --------
    TiffStack_libtiff, TiffStack_tiffile, ImageSequence
    """
    def __init__(self, fname):

        self.im = Image.open(fname)
        self._filename = fname  # used by __repr__

        self.im.seek(0)
        # this will need some work to deal with color
        res = self.im.tag[0x102][0]
        self._dtype = _dtype_map.get(res, np.int16)
        try:
            samples_per_pixel = self.im.tag[0x115][0]
            if samples_per_pixel != 1:
                raise ValueError("support for color not implemented")
        except:
            pass
        # get image dimensions from the meta data the order is flipped
        # due to row major v col major ordering in tiffs and numpy
        w = self.im.tag[0x101][0]
        h = self.im.tag[0x100][0]
        samples_per_px = self.im.tag[0x115][0]
        if samples_per_px != 1:
            self._im_sz = (w, h, samples_per_px)
        else:
            self._im_sz = (w, h)
        # walk through stack to get length, there has to
        # be a better way to do this
        for j in itertools.count():
            try:
                self.im.seek(j)
            except EOFError:
                break

        self._count = j
        self.cur = self.im.tell()

    def get_frame(self, j):
        '''Extracts the jth frame from the image sequence.
        if the frame does not exist return None'''
        # PIL does not support random access. If we need to rewind, re-open
        # the file.
        if j < self.cur:
            self.im.close()
            self.im = Image.open(self._filename)
            self.im.seek(j)
        elif j > self.cur:
            self.im.seek(j)
        elif j > len(self):
            raise IndexError("out of bounds; length is {0}".format(len(self)))
        # If j == self.cur, do nothing.
        self.cur = self.im.tell()
        res = np.reshape(self.im.getdata(), self._im_sz)
        return Frame(res, frame_no=j, metadata=self._read_metadata())

    def _read_metadata(self):
        """Read metadata for current frame and return as dict"""
        try:
            tags = self.im.tag_v2  # for Pillow >= v3.0.0
        except AttributeError:
            tags = self.im.tag  # for Pillow < v3.0.0
        md = {}
        try:
            md["ImageDescription"] = tags[270]
        except KeyError:
            pass
        try:
            md["DateTime"] = _tiff_datetime(tags[306])
        except KeyError:
            pass
        try:
            md["Software"] = tags[305]
        except KeyError:
            pass
        try:
            md["DocumentName"] = tags[269]
        except KeyError:
            pass
        return md

    @property
    def pixel_type(self):
        return self._dtype

    @property
    def frame_shape(self):
        return self._im_sz

    def __len__(self):
        return self._count

    def close(self):
        self.im.close()
        super().close()

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {w} x {h}
Pixel Datatype: {dtype}""".format(w=self.frame_shape[0],
                                  h=self.frame_shape[1],
                                  count=len(self),
                                  filename=self._filename,
                                  dtype=self.pixel_type)


class MM_TiffStack(TiffStack_pil):
    """
    Specialized class for dealing with meta-morph tiffs.

    A function `get_meta` is added which extracts and parses
    the xml meta-data field.
    """
    def get_meta(self, j):
        cur = self.im.tell()
        if cur != j:
            self.im.seek(j)
            xml_str = im_wrap.im.tag[270]
            self.im.seek(cur)
        else:
            xml_str = im_wrap.im.tag[270]
        return _parse_mm_xml_string(xml_str)


class TiffSeries(FramesSequence):
    '''
    Class for dealing with a series of tiffs which are systematically
    named.

    Parameters
    ----------

    name_template : string
      input string for `format` to generate the file names keyed on `ind`.  For
      example for the set of files ('/home/user/data/f_001.tif',
       '/home/user/data/f_002.tif', ...) the correct input would be
        `name_template=/home/user/data/f_{ind:03d}.tif'

    offset : int, default 1
        The file number for frame 0.

    '''
    def __init__(self, name_template, offset=1):

        self._name_template = name_template
        self._offset = offset

        im = Image.open(self._name_template.format(ind=self._offset))
        # get image dimensions from the meta data the order is flipped
        # due to row major v col major ordering in tiffs and numpy
        self._im_sz = (im.tag[0x101][0],
                       im.tag[0x100][0])

        res = im.tag[0x102][0]
        self._dtype = _dtype_map.get(res, np.int16)

        try:
            samples_per_pixel = im.tag[0x115][0]
            if samples_per_pixel != 1:
                raise ValueError("support for color not implemented")
        except:
            pass

        j = 0
        # sort out how many there are
        while os.path.isfile(name_template.format(j + offset)):
            j += 1

        self._count = j

    def get_frame(self, j):
        '''Extracts the jth frame from the image sequence.
        if the frame does not exist return None'''

        im = Image.open(self._name_template.format(ind=j + self._offset))

        return np.reshape(im.getdata(), self._im_sz)

    @property
    def pixel_type(self):
        return self._dtype

    @property
    def frame_shape(self):
        return self._im_sz

    def __len__(self):
        return self._count

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Source: {name_template}
Length: {count} frames
Frame Shape: {w} x {h}
Pixel Datatype: {dtype}""".format(w=self.frame_shape[0],
                                  h=self.frame_shape[1],
                                  count=len(self),
                                  name_template=self._name_template,
                                  dtype=self.pixel_type)


# needed for the wrapper classes
def _parse_mm_xml_string(xml_str):
    """
    Parses Meta-Morph xml meta-data strings to a dictionary

    Parameters
    ----------
    xml_str : string
        A properly formed xml string

    Returns
    -------
    f : dict
        A dictionary object containing the meta-data
    """
    def _write(md_dict, name, val):
        if (name == "acquisition-time-local"
             or name == "modification-time-local"):
            tmp = int(val[18:])
            val = val[:18] + "%(#)03d" % {"#": tmp}
            val = datetime.datetime.strptime(val,
                                             '%Y%m%d %H:%M:%S.%f')
        md_dict[name] = val

    def _parse_attr(file_obj, dom_obj):
        if dom_obj.getAttribute("id") == "Description":
            _parse_des(file_obj, dom_obj)
        elif dom_obj.getAttribute("type") == "int":
            _write(file_obj, dom_obj.getAttribute("id"),
                   int(dom_obj.getAttribute("value")))
        elif dom_obj.getAttribute("type") == "float":
            _write(file_obj, dom_obj.getAttribute("id"),
                   float(dom_obj.getAttribute("value")))
        else:
            _write(file_obj, dom_obj.getAttribute("id"),
                   dom_obj.getAttribute("value").encode('ascii'))

    def _parse_des(file_obj, des_obj):
        des_string = des_obj.getAttribute("value")
        des_split = des_string.split("&#13;&#10;")

        for x in des_split:
            tmp_split = x.split(":")
            if len(tmp_split) == 2:
                _write(file_obj, tmp_split[0],
                       tmp_split[1].encode('ascii'))

    dom = xml.dom.minidom.parseString(xml_str)

    props = dom.getElementsByTagName("prop")
    f = dict()
    for p in props:
        _parse_attr(f, p)

    return f



from pims.base_frames import FramesSequence
from pims.frame import Frame
import warnings

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        from moviepy.editor import VideoFileClip
except ImportError:
    VideoFileClip = None
except RuntimeError:
    # there is an incompatibility between moviepy 2.3.5 and imageio >= 2.5.0
    VideoFileClip = None


def available():
    return VideoFileClip is not None


class MoviePyReader(FramesSequence):
    """
    Read images from a video file via a MoviePy interface.
    """
    class_priority = 4
    @classmethod
    def class_exts(cls):
        return {'mov', 'mp4', 'avi', 'mpeg', 'wmv', 'mkv'}
    def __init__(self, filename):
        if VideoFileClip is None:
            raise ImportError('The MoviePyReader requires moviepy to work.')
        self.clip = VideoFileClip(filename)
        self.filename = filename
        self._fps = self.clip.fps
        self._len = int(self.clip.fps * self.clip.end)

        first_frame = self.clip.get_frame(0)
        self._shape = first_frame.shape
        self._dtype = first_frame.dtype

    def get_frame(self, i):
        return Frame(self.clip.get_frame(i / self._fps), frame_no=i)

    def __len__(self):
        return self._len

    @property
    def frame_shape(self):
        return self._shape

    @property
    def frame_rate(self):
        return self._fps

    @property
    def pixel_type(self):
        return self._dtype  



import unittest
from itertools import chain, permutations
import numpy as np
from numpy.testing import assert_equal
from pims import FramesSequenceND, Frame


class TestMultidimensional(unittest.TestCase):
    def setUp(self):
        class IndexReturningReader(FramesSequenceND):
            @property
            def pixel_type(self):
                return np.uint8

            def __init__(self, c, m, t, z):
                super(IndexReturningReader, self).__init__()
                self._init_axis('x', 6)
                self._init_axis('y', 1)
                self._init_axis('z', z)
                self._init_axis('c', c)
                self._init_axis('m', m)
                self._init_axis('t', t)
                self.bundle_axes = 'yx'

            def get_frame_2D(self, **ind):
                result_order = ['c', 'm', 't', 'x', 'y', 'z']
                result = [ind[a] for a in result_order]
                metadata = {i: ind[i] for i in result_order}
                return Frame([result], metadata=metadata).astype(np.uint8)

        self.v = IndexReturningReader(c=3, m=5, t=100, z=20)

    def test_iterate(self):
        # index order c, m, t, x, y, z
        self.v.iter_axes = 't'
        for i in [0, 1, 15]:
            assert_equal(self.v[i], [[0, 0, i, 0, 0, 0]])
        self.v.iter_axes = 'm'
        for i in [0, 1, 3]:
            assert_equal(self.v[i], [[0, i, 0, 0, 0, 0]])
        self.v.iter_axes = 'zc'
        assert_equal(self.v[0], [[0, 0, 0, 0, 0, 0]])
        assert_equal(self.v[2], [[2, 0, 0, 0, 0, 0]])
        assert_equal(self.v[30], [[0, 0, 0, 0, 0, 10]])
        self.v.iter_axes = 'cz'
        assert_equal(self.v[0], [[0, 0, 0, 0, 0, 0]])
        assert_equal(self.v[4], [[0, 0, 0, 0, 0, 4]])
        assert_equal(self.v[21], [[1, 0, 0, 0, 0, 1]])
        self.v.iter_axes = 'tzc'
        assert_equal(self.v[0], [[0, 0, 0, 0, 0, 0]])
        assert_equal(self.v[4], [[1, 0, 0, 0, 0, 1]])
        assert_equal(self.v[180], [[0, 0, 3, 0, 0, 0]])
        assert_equal(self.v[210], [[0, 0, 3, 0, 0, 10]])
        assert_equal(self.v[212], [[2, 0, 3, 0, 0, 10]])

    def test_default(self):
        self.v.iter_axes = 't'
        self.v.default_coords['m'] = 2
        for i in [0, 1, 3]:
            assert_equal(self.v[i], [[0, 2, i, 0, 0, 0]])
        self.v.default_coords['m'] = 0
        for i in [0, 1, 3]:
            assert_equal(self.v[i], [[0, 0, i, 0, 0, 0]])

    def test_bundle(self):
        self.v.bundle_axes = 'zyx'
        assert_equal(self.v[0].shape, (20, 1, 6))
        self.v.bundle_axes = 'cyx'
        assert_equal(self.v[0].shape, (3, 1, 6))
        self.v.bundle_axes = 'czyx'
        assert_equal(self.v[0].shape, (3, 20, 1, 6))
        self.v.bundle_axes = 'zcyx'
        assert_equal(self.v[0].shape, (20, 3, 1, 6))

    def test_frame_no(self):
        self.v.iter_axes = 't'
        for i in np.random.randint(0, 100, 10):
            assert_equal(self.v[i].frame_no, i)
        self.v.iter_axes = 'zc'
        for i in np.random.randint(0, 3*20, 10):
            assert_equal(self.v[i].frame_no, i)

    def test_metadata(self):
        self.v.iter_axes = 't'
        self.v.bundle_axes = 'czyx'
        md = self.v[15].metadata

        assert_equal(md['axes'], tuple('czyx'))
        assert_equal(md['coords']['t'], 15)
        assert_equal(md['coords']['m'], 0)
        for ax in 'czyx':
            assert ax not in md['coords']

        # if metadata is provided, it should have the correct shape
        assert_equal(md['z'].shape, (3, 20))  # shape 'c', 'z'
        assert_equal(md['z'][:, 5], 5)
        assert_equal(md['c'][1, :], 1)

        # if a metadata field is equal for all frames, it should be a scalar
        assert_equal(md['t'], 15)

    def test_mutability(self):
        # test for issues that may arise when properties return mutable objects

        # the list bundle_axes
        self.v.bundle_axes = ['z', 'y', 'x']
        temp = self.v.bundle_axes
        temp = []
        assert_equal(self.v[0].shape, (20, 1, 6))

        # the list iter_axes
        self.v.iter_axes = ['t']
        temp = self.v.iter_axes
        temp = []
        assert_equal(len(self.v), 100)

        # the dict default_coords
        # changing an nonexisting item on a copy does nothing
        a = dict(self.v.default_coords)
        a['non_existing'] = 0
        # but assigning it again raises
        with self.assertRaises(ValueError) as cm:
            self.v.default_coords = a

        # changing an nonexisting item should raise
        with self.assertRaises(ValueError) as cm:
            self.v.default_coords['non_existing'] = 0


class RandomReaderFlexible(FramesSequenceND):
    def __init__(self, reads_axes, **sizes):
        super(RandomReaderFlexible, self).__init__()
        for key in sizes:
            self._init_axis(key, sizes[key])
        self._gf_shape = tuple([sizes[a] for a in reads_axes])
        self._register_get_frame(self._get_frame, reads_axes)

    def _get_frame(self, **ind):
        return np.empty(self._gf_shape, dtype=np.uint8)

    @property
    def pixel_type(self):
        return np.uint8


class TestFramesSequenceND(unittest.TestCase):
    def test_flexible_get_frame(self):
        sizes = dict(x=128, y=64, c=3, z=10)
        all_modes = list(chain(*[permutations(sizes, x)
                               for x in range(1, len(sizes) + 1)]))
        for read_mode in all_modes:
            reader = RandomReaderFlexible(read_mode, **sizes)
            for bundle in all_modes:
                reader.bundle_axes = bundle
                assert_equal(reader[0].shape, [sizes[k] for k in bundle])

    def test_flexible_get_frame_compatibility(self):
        class RandomReader_2D(FramesSequenceND):
            def __init__(self, **sizes):
                super(RandomReader_2D, self).__init__()
                for key in sizes:
                    self._init_axis(key, sizes[key])

            def get_frame_2D(self, **ind):
                return np.random.randint(0, 255, (64, 128)).astype(np.uint8)

            @property
            def pixel_type(self):
                return np.uint8

        sizes = dict(x=128, y=64, c=3, z=10)
        all_modes = chain(*[permutations(sizes, x)
                            for x in range(1, len(sizes) + 1)])
        reader = RandomReader_2D(**sizes)
        for bundle in all_modes:
            reader.bundle_axes = bundle
            assert_equal(reader[0].shape, [sizes[k] for k in bundle])

if __name__ == '__main__':
    unittest.main()



import numpy as np
from slicerator import pipeline, Pipeline


@pipeline
def as_grey(frame):
    """Convert a 2D image or PIMS reader to greyscale.

    This weights the color channels according to their typical
    response to white light.

    It does nothing if the input is already greyscale.
    """
    if len(frame.shape) == 2:
        return frame
    else:
        red = frame[:, :, 0]
        green = frame[:, :, 1]
        blue = frame[:, :, 2]
        return 0.2125 * red + 0.7154 * green + 0.0721 * blue

# "Gray" is the more common spelling
as_gray = as_grey

# Source of this patch: https://github.com/scikit-image/scikit-image/pull/3556
# See also: https://github.com/numpy/numpy/pull/11966

from numpy.lib.arraypad import _as_pairs

def validate_lengths(ar, crop_width):
    return _as_pairs(crop_width, ar.ndim, as_index=True)

def _crop(frame, bbox):
    return frame[bbox[0]:bbox[2], bbox[1]:bbox[3]]


@pipeline
class crop(Pipeline):
    """Crop image or image-reader`reader` by `crop_width` along each dimension.

    Parameters
    ----------
    ar : array-like of rank N
       Input array.
    crop_width : {sequence, int}
       Number of values to remove from the edges of each axis.
       ``((before_1, after_1),`` ... ``(before_N, after_N))`` specifies
       unique crop widths at the start and end of each axis.
       ``((before, after),)`` specifies a fixed start and end crop
       for every axis.
       ``(n,)`` or ``n`` for integer ``n`` is a shortcut for
       before = after = ``n`` for all axes.
    order : {'C', 'F', 'A', 'K'}, optional
       control the memory layout of the copy. See ``np.copy``.
    Returns
    -------
    cropped : array
       The cropped array.

    See Also
    --------
    Source: ``skimage.util.crop`` (v0.12.3)
    """
    def __init__(self, reader, crop_width, order='K'):
        # We have to know the frame shape that is returned by the reader.
        try:  # In case the reader is a FramesSequence, there is an attribute
            shape = reader.frame_shape
            first_frame = np.empty(shape, dtype=bool)
        except AttributeError:
            first_frame = reader[0]
            shape = first_frame.shape
        # Validate the crop widths on the first frame
        crops = validate_lengths(first_frame, crop_width)
        self._crop_slices = tuple([slice(a, shape[i] - b)
                             for i, (a, b) in enumerate(crops)])
        self._crop_shape = tuple([shape[i] - b - a
                                  for i, (a, b) in enumerate(crops)])
        self._crop_order = order
        # We could pass _crop to proc_func. However this adds an extra copy
        # operation. Therefore we define our own here.
        super(self.__class__, self).__init__(None, reader)

    def _get(self, key):
        ar = self._ancestors[0][key]
        return np.array(ar[self._crop_slices], order=self._crop_order,
                        copy=True)

    @property
    def frame_shape(self):
        return self._crop_shape



# Tests for norpix_reader.py

import os
from datetime import datetime
import unittest
import numpy as np
import pims
import pims.norpix_reader

tests_path, _ = os.path.split(os.path.abspath(__file__))

class _common_norpix_sample_tests(object):
    """Test the Norpix .seq reader on a sample file."""
    def setUp(self):
        self.seq = pims.open(self.sample_filename, **self.options)

    def tearDown(self):
        if hasattr(self, 'seq'):
            self.seq.close()

    def test_type(self):
        assert isinstance(self.seq, pims.norpix_reader.NorpixSeq)

    def test_metadata(self):
        s = self.seq
        assert isinstance(len(s), int)
        assert np.issubdtype(s.pixel_type, np.number)
        assert s.width > 0
        assert s.height > 0
        assert len(s.filename)
        assert len(repr(s))
        for k in ('description', 'bit_depth_real', 'origin',
                  'suggested_frame_rate', 'width', 'height',
                  'gamut'):
            assert k in s.metadata

    def test_get_frame(self):
        s = self.seq
        hashes = set()  # Check that each frame is unique
        for i in range(len(s)):
            fr = s[i]
            fhash = hash(np.array(fr).tobytes())
            assert fhash not in hashes
            hashes.add(fhash)

            for k in ('time', 'time_float', 'gamut'):
                assert k in fr.metadata

            assert fr.shape[1] == s.width
            assert fr.shape[0] == s.height

    def test_get_time(self):
        """Check all 3 ways to get time of a frame."""
        s = self.seq
        maxtime = 0.
        for i in range(len(s)):
            tdt = s.get_time(i)
            tfloat = s.get_time_float(i)
            fr = s[i]

            assert isinstance(tdt, datetime)

            assert tfloat > maxtime
            maxtime = tfloat

            assert fr.metadata['time'] == tdt
            assert fr.metadata['time_float'] == tfloat

    def test_get_time_mapped(self):
        s = self.seq
        sli = s[4:]
        assert s.get_time(4) == sli.get_time(0)
        assert s.get_time_float(4) == list(sli.get_time_float[:])[0]

    def test_dump_times(self):
        assert isinstance(self.seq.dump_times_float(), np.ndarray)

    def test_repr(self):
        assert len(repr(self.seq))


class test_norpix5_sample(_common_norpix_sample_tests, unittest.TestCase):
    def setUp(self):
        self.sample_filename = os.path.join(tests_path, 'data',
                                            'sample_norpix5.seq')
        if not os.path.exists(self.sample_filename):
            raise unittest.SkipTest('(Large) Norpix v5 sample file not found. '
                                    'Skipping.')

        self.options = {}
        super(test_norpix5_sample, self).setUp()


class _norpix6_sample_tests(_common_norpix_sample_tests):
    def setUp(self):
        self.sample_filename = os.path.join(tests_path, 'data',
                                            'sample_norpix6.seq')
        super(_norpix6_sample_tests, self).setUp()

    def test_specific_file_info(self):
        """Tests based on the specific file in the repo."""
        s = self.seq
        assert len(s) == 6
        assert s.height == 32
        assert s.width == 36


class test_defaults(_norpix6_sample_tests, unittest.TestCase):
    def setUp(self):
        self.options = {}
        super(test_defaults, self).setUp()

    def test_specific_file_dtype(self):
        """Based on the specific file in the repo."""
        assert self.seq[0].dtype == np.uint8


# class test_dtype(_norpix6_sample_tests, unittest.TestCase):
#     def setUp(self):
#         self.options = {}
#         self.dtype = np.float_
#         self.options['dtype'] = self.dtype
#         super(test_dtype, self).setUp()
#
#     def test_dtype(self):
#         fr = self.seq[0]
#         assert fr.dtype == self.dtype
#
#
# class test_process_func(_norpix6_sample_tests, unittest.TestCase):
#     def setUp(self):
#         self.options = {}
#         self.options['dtype'] = np.float_
#         self.options['process_func'] = lambda x: -x
#         super(test_process_func, self).setUp()
#
#     def test_process_func(self):
#         fr = self.seq[0]
#         assert np.all(fr <= 0)


class test_as_raw(_norpix6_sample_tests, unittest.TestCase):
    def setUp(self):
        self.options = {'as_raw': True}
        super(test_as_raw, self).setUp()
    #
    # def test_post_hoc_process_func(self):
    #     testbyte = self.seq[0][0,0]
    #     self.seq.set_process_func(lambda x: 255 - x)
    #     assert self.seq[0][0,0] == 255 - testbyte



import os
import glob
import fnmatch
from warnings import warn
import re
import zipfile
from io import BytesIO
from functools import partial

import numpy as np

import pims
from pims.base_frames import FramesSequence, FramesSequenceND
from pims.frame import Frame
from pims.image_reader import imread
from pims.utils.sort import natural_keys


class ImageSequence(FramesSequence):
    """Read a directory of sequentially numbered image files into an
    iterable that returns images as numpy arrays.

    Parameters
    ----------
    path_spec : string or iterable of strings
        a directory or, safer, a pattern like path/to/images/*.png
        which will ignore extraneous files or a list of files to open
        in the order they should be loaded. When a path to a zipfile is
        specified, all files in the zipfile will be loaded.
    plugin : string
        Passed on to skimage.io.imread if scikit-image is available.
        If scikit-image is not available, this will be ignored and a warning
        will be issued. Not available in combination with zipfiles.

    Examples
    --------
    >>> video = ImageSequence('path/to/images/*.png')  # or *.tif, or *.jpg
    >>> imshow(video[0]) # Show the first frame.
    >>> imshow(video[-1]) # Show the last frame.
    >>> imshow(video[1][0:10, 0:10]) # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video
    """
    def __init__(self, path_spec, plugin=None):
        if not imread.__module__.startswith("skimage"):
            if plugin is not None:
                warn("A plugin was specified but ignored. Plugins can only "
                     "be specified if scikit-image is available. Instead, "
                     "ImageSequence will use imageio")
            self.kwargs = dict()
        else:
            self.kwargs = dict(plugin=plugin)

        self._is_zipfile = False
        self._zipfile = None
        self._get_files(path_spec)

        tmp = self.imread(self._filepaths[0], **self.kwargs)
        self._first_frame_shape = tmp.shape
        self._dtype = tmp.dtype

    def close(self):
        if self._is_zipfile:
            self._zipfile.close()
        super(ImageSequence, self).close()

    def __del__(self):
        self.close()

    def imread(self, filename, **kwargs):
        if self._is_zipfile:
            file_handle = BytesIO(self._zipfile.read(filename))
            return imread(file_handle, **kwargs)
        else:
            return imread(filename, **kwargs)

    def _get_files(self, path_spec):
        # deal with if input is _not_ a string
        if not isinstance(path_spec, str):
            # assume it is iterable and off we go!
            self._filepaths = list(path_spec)
            self._count = len(self._filepaths)
            return

        if zipfile.is_zipfile(path_spec):
            self._is_zipfile = True
            self.pathname = os.path.abspath(path_spec)
            self._zipfile = zipfile.ZipFile(path_spec, 'r')
            filepaths = [fn for fn in self._zipfile.namelist()
                         if fnmatch.fnmatch(fn, '*.*')]
            self._filepaths = sorted(filepaths, key=natural_keys)
            self._count = len(self._filepaths)
            if 'plugin' in self.kwargs and self.kwargs['plugin'] is not None:
                warn("A plugin cannot be combined with reading from an "
                     "archive. Extract it if you want to use the plugin.")
            return

        self.pathname = os.path.abspath(path_spec)  # used by __repr__
        if os.path.isdir(path_spec):
            warn("Loading ALL files in this directory. To ignore extraneous "
                 "files, use a pattern like 'path/to/images/*.png'",
                 UserWarning)
            directory = path_spec
            filenames = os.listdir(directory)
            make_full_path = lambda filename: (
                os.path.abspath(os.path.join(directory, filename)))
            filepaths = list(map(make_full_path, filenames))
        else:
            filepaths = glob.glob(path_spec)
        self._filepaths = sorted(filepaths, key=natural_keys)
        self._count = len(self._filepaths)

        # If there were no matches, this was probably a user typo.
        if self._count == 0:
            raise IOError("No files were found matching that path.")

    def get_frame(self, j):
        if j > self._count:
            raise ValueError("File does not contain this many frames")
        res = self.imread(self._filepaths[j], **self.kwargs)
        return Frame(res, frame_no=j)

    def __len__(self):
        return self._count

    @property
    def frame_shape(self):
        return self._first_frame_shape

    @property
    def pixel_type(self):
        return self._dtype

    def __repr__(self):
        # May be overwritten by subclasses
        try:
            source = self.pathname
        except AttributeError:
            source = '(list of images)'
        return """<Frames>
Source: {pathname}
Length: {count} frames
Frame Shape: {frame_shape!r}
Pixel Datatype: {dtype}""".format(frame_shape=self.frame_shape,
                                  count=len(self),
                                  pathname=source,
                                  dtype=self.pixel_type)


def filename_to_indices(filename, identifiers='tzc'):
    """ Find ocurrences of axis indices (e.g. t001, z06, c2)
    in a filename and returns a list of indices.

    Parameters
    ----------
    filename : string
        filename to be searched for indices
    identifiers : string or list of strings, optional
        iterable of N strings preceding axis indices, in that order

    Returns
    ---------
    list of int
        axis indices. Elements default to 0 when index was not found.

    """
    escaped = [re.escape(a) for a in identifiers]
    axes = re.findall('(' + '|'.join(escaped) + r')(\d+)',
                            filename)
    if len(axes) > len(identifiers):
        axes = axes[-3:]
    order = [a[0] for a in axes]
    result = [0] * len(identifiers)
    for (i, col) in enumerate(identifiers):
        try:
            result[i] = int(axes[order.index(col)][1])
        except ValueError:
            result[i] = 0
    return result


class ReaderSequence(FramesSequenceND):
    """Construct a reader from a directory of ND image files.

    Parameters
    ----------
    path_spec : string or iterable of strings
        a directory or, safer, a pattern like path/to/images/*.png
        which will ignore extraneous files or a list of files to open
        in the order they should be loaded. When a path to a zipfile is
        specified, all files in the zipfile will be loaded. The filenames
        should contain the indices of T, Z and C, preceded by a axis
        identifier such as: 'file_t001c05z32'.
    axis_name : string, optional
        The name of the added axis. Default 't'.
    """
    def __init__(self, path_spec, reader_cls=None, axis_name='t', **kwargs):
        FramesSequenceND.__init__(self)

        self.kwargs = kwargs
        if reader_cls is None:
            self.reader_cls = pims.open
        else:
            self.reader_cls = reader_cls
        self._get_files(path_spec)

        with self.reader_cls(self._filepaths[0], **self.kwargs) as reader:
            if not isinstance(reader, FramesSequenceND):
                raise ValueError("Reader is not subclass of FramesSequenceND")
            for ax in reader.axes:
                self._init_axis(ax, reader.sizes[ax])
            self._pixel_type = reader.pixel_type
        self._imseq_axis = axis_name
        self._init_axis(axis_name, self._count)
        self.iter_axes = [axis_name]

    @property
    def bundle_axes(self):
        return self._bundle_axes[:]

    @bundle_axes.setter
    def bundle_axes(self, value):
        """Overrides the baseclass 'smart' bundle_axes, as _get_frame_wrapped
        uses the child reader bundle axes logic."""
        value = list(value)
        invalid = [k for k in value if k not in self._sizes]
        if invalid:
            raise ValueError("axes %r do not exist" % invalid)

        if self._imseq_axis in self.bundle_axes:
            raise ValueError('The sequence axis cannot be bundled.')
        for k in value:
            if k in self._iter_axes:
                del self._iter_axes[self._iter_axes.index(k)]
        self._bundle_axes = value
        self._get_frame_wrapped = self._get_seq_frame

    def _get_seq_frame(self, **coords):
        i = coords.pop(self._imseq_axis)
        with self.reader_cls(self._filepaths[i], **self.kwargs) as reader:
            # check whether the reader has the expected shape
            for ax in self.sizes:
                if ax == self._imseq_axis:
                    continue
                if ax not in reader.sizes:
                    raise RuntimeError('{} does not have '
                                       'axis {}'.format(self._filepaths[i], ax))
                if reader.sizes[ax] != self.sizes[ax]:
                    raise RuntimeError('In {}, the size of axis {} was unexpect'
                                       'ed'.format(self._filepaths[i], ax))
            reader.bundle_axes = self.bundle_axes
            result = reader._get_frame_wrapped(**coords)
        return result

    @property
    def pixel_type(self):
        return self._pixel_type

    def __repr__(self):
        try:
            source = self.pathname
        except AttributeError:
            source = '(list of images)'
        s = "<ReaderSequence>\nSource: {0}\n".format(source)
        s += "Axes: {0}\n".format(self.ndim)
        for dim in self._sizes:
            s += "Axis '{0}' size: {1}\n".format(dim, self._sizes[dim])
        s += """Pixel Datatype: {dtype}""".format(dtype=self.pixel_type)
        return s

    def _get_files(self, path_spec):
        # deal with if input is _not_ a string
        if not isinstance(path_spec, str):
            # assume it is iterable and off we go!
            self._filepaths = sorted(list(path_spec), key=natural_keys)
            self._count = len(path_spec)
            return

        if zipfile.is_zipfile(path_spec):
            self._is_zipfile = True
            self.pathname = os.path.abspath(path_spec)
            self._zipfile = zipfile.ZipFile(path_spec, 'r')
            filepaths = [fn for fn in self._zipfile.namelist()
                         if fnmatch.fnmatch(fn, '*.*')]
            self._filepaths = sorted(filepaths, key=natural_keys)
            self._count = len(self._filepaths)
            if 'plugin' in self.kwargs and self.kwargs['plugin'] is not None:
                warn("A plugin cannot be combined with reading from an "
                     "archive. Extract it if you want to use the plugin.")
            return

        self.pathname = os.path.abspath(path_spec)  # used by __repr__
        if os.path.isdir(path_spec):
            warn("Loading ALL files in this directory. To ignore extraneous "
                 "files, use a pattern like 'path/to/images/*.png'",
                 UserWarning)
            directory = path_spec
            filenames = os.listdir(directory)
            make_full_path = lambda filename: (
                os.path.abspath(os.path.join(directory, filename)))
            filepaths = list(map(make_full_path, filenames))
        else:
            filepaths = glob.glob(path_spec)
        self._filepaths = sorted(filepaths, key=natural_keys)
        self._count = len(self._filepaths)

        # If there were no matches, this was probably a user typo.
        if self._count == 0:
            raise IOError("No files were found matching that path.")


class ImageSequenceND(FramesSequenceND, ImageSequence):
    """Read a directory of multi-indexed image files into an iterable that
    returns images as numpy arrays. By default, the extra axes are
    denoted with t, z, c.

    Parameters
    ----------
    path_spec : string or iterable of strings
        a directory or, safer, a pattern like path/to/images/*.png
        which will ignore extraneous files or a list of files to open
        in the order they should be loaded. When a path to a zipfile is
        specified, all files in the zipfile will be loaded. The filenames
        should contain the indices of T, Z and C, preceded by a axis
        identifier such as: 'file_t001c05z32'.
    plugin : string, optional
        Passed on to skimage.io.imread if scikit-image is available.
        If scikit-image is not available, this will be ignored and a warning
        will be issued. Not available in combination with zipfiles.
    axes_identifiers : iterable of strings, optional
        N strings preceding axes indices. Default 'tzc'. x and y are not
        allowed. c is not allowed when images are RGB.

    Attributes
    ----------
    axes : list of strings
        List of all available axes
    ndim : int
        Number of image axes
    sizes : dict of int
        Dictionary with all axis sizes
    frame_shape : tuple of int
        Shape of frames that will be returned by get_frame
    iter_axes : iterable of strings
        This determines which axes will be iterated over by the FramesSequence.
        The last element in will iterate fastest. x and y are not allowed.
    bundle_axes : iterable of strings
        This determines which axes will be bundled into one Frame. The axes in
        the ndarray that is returned by get_frame have the same order as the
        order in this list. The last two elements have to be ['y', 'x'].
        If the 'z' axis exists then it defaults to ['z', 'y', 'x']
    default_coords : dict of int
        When an axis is not present in both iter_axes and bundle_axes, the
        coordinate contained in this dictionary will be used.
    is_rgb : boolean
        True when the input image is an RGB image.
    is_interleaved : boolean
        Applicable to RGB images. Signifies the position of the rgb axis in
        the input image. True when color data is stored in the last dimension.
    """
    def __init__(self, path_spec, plugin=None, axes_identifiers='tzc'):
        FramesSequenceND.__init__(self)
        if 'x' in axes_identifiers:
            raise ValueError("Axis 'x' is reserved")
        if 'y' in axes_identifiers:
            raise ValueError("Axis 'y' is reserved")
        self.axes_identifiers = axes_identifiers
        ImageSequence.__init__(self, path_spec, plugin)
        shape = self._first_frame_shape
        if len(shape) == 2:
            self._init_axis('y', shape[0])
            self._init_axis('x', shape[1])
            self._register_get_frame(self.get_frame_2D, 'yx')
            self.is_rgb = False
        elif len(shape) == 3 and shape[2] in [3, 4]:
            self._init_axis('y', shape[0])
            self._init_axis('x', shape[1])
            self._init_axis('c', shape[2])
            self._register_get_frame(self.get_frame_2D, 'yxc')
            self.is_rgb = True
            self.is_interleaved = True
        elif len(shape) == 3:
            self._init_axis('c', shape[0])
            self._init_axis('y', shape[1])
            self._init_axis('x', shape[2])
            self._register_get_frame(self.get_frame_2D, 'cyx')
            self.is_rgb = True
            self.is_interleaved = False
        else:
            raise IOError("Could not interpret image shape.")

        if self.is_rgb and 'c' in self.axes_identifiers:
            raise ValueError("Axis identifier 'c' is reserved when "
                             "images are rgb.")

        if 't' in self.axes:
            self.iter_axes = ['t']  # iterate over t
        if 'z' in self.axes:
            self.bundle_axes = ['z', 'y', 'x']  # return z-stacks

    def _get_files(self, path_spec):
        super(ImageSequenceND, self)._get_files(path_spec)
        self._toc = np.array([filename_to_indices(f, self.axes_identifiers)
                              for f in self._filepaths])
        for n, name in enumerate(self.axes_identifiers):
            if np.all(self._toc[:, n] == 0):
                self._toc = np.delete(self._toc, n, axis=1)
            else:
                self._toc[:, n] = self._toc[:, n] - min(self._toc[:, n])
                self._init_axis(name, max(self._toc[:, n]) + 1)
        self._filepaths = np.array(self._filepaths)

    def get_frame(self, i):
        frame = super(ImageSequenceND, self).get_frame(i)
        return Frame(frame, frame_no=i)

    def get_frame_2D(self, **ind):
        if self.is_rgb:
            c = ind['c']
            row = [ind[name] for name in self.axes_identifiers if name != 'c']
        else:
            row = [ind[name] for name in self.axes_identifiers]
        i = np.argwhere(np.all(self._toc == row, 1))[0, 0]
        return self.imread(self._filepaths[i], **self.kwargs)

    def __repr__(self):
        try:
            source = self.pathname
        except AttributeError:
            source = '(list of images)'
        s = "<ImageSequenceND>\nSource: {0}\n".format(source)
        s += "Axes: {0}\n".format(self.ndim)
        for dim in self._sizes:
            s += "Axis '{0}' size: {1}\n".format(dim, self._sizes[dim])
        s += """Pixel Datatype: {dtype}""".format(dtype=self.pixel_type)
        return s


def customize_image_sequence(imread_func, name=None):
    """Class factory for ImageSequence with customized image reader.

    Parameters
    ----------
    imread_func : callable
        image reader
    name : str or None
        name of class returned; if None, 'CustomImageSequence' is used.

    Returns
    -------
    type : a subclass of ImageSequence
        This subclass has its image-opening method, imread, overriden
        by the passed function.

    Example
    -------
    >>> # my_func accepts a filename and returns a numpy array
    >>> MyImageSequence = customize_image_sequence(my_func)
    >>> frames = MyImageSequence('path/to/my_weird_files*')
    """
    class CustomImageSequence(ImageSequence):
        def imread(self, filename, **kwargs):
            return imread_func(filename, **kwargs)
    if name is not None:
        CustomImageSequence.__name__ = name
    return CustomImageSequence



import unittest
import numpy as np
from pims.frame import Frame


def _skip_if_no_jinja2():
    try:
        import jinja2
    except ImportError:
        raise unittest.SkipTest('Jinja2 not installed. Skipping.')


def test_scalar_casting():
    tt = Frame(np.ones((5, 3)), frame_no=42)
    sum1 = tt.sum()
    assert np.isscalar(sum1)
    sum2 = tt.sum(keepdims=True)
    assert sum2.ndim == 2
    assert sum2.frame_no == tt.frame_no


def test_creation_md():
    md_dict = {'a': 1}
    frame_no = 42
    tt = Frame(np.ones((5, 3)), frame_no=frame_no, metadata=md_dict)
    assert tt.metadata == md_dict
    assert tt.frame_no == frame_no


def test_repr_html_():
    _skip_if_no_jinja2()
    # This confims a bugfix, where 16-bit images would raise
    # an error.
    Frame(10000*np.ones((50, 50), dtype=np.uint16))._repr_html_()


def test_copy():
    md_dict = {'a': 1}
    frame_no = 42
    tt_base = Frame(np.ones((5, 3)), frame_no=frame_no, metadata=md_dict)
    tt = Frame(tt_base)
    assert tt.metadata == md_dict
    assert tt.frame_no == frame_no


def test_copy_override_frame():
    frame_no = 42
    tt_base = Frame(np.ones((5, 3)), frame_no=frame_no)
    frame_no_2 = 123
    tt = Frame(tt_base, frame_no=frame_no_2)
    assert tt.frame_no == frame_no_2


def test_copy_update_md():
    frame_no = 42
    md_dict = {'a': 1}
    md_dict2 = {'b': 1}
    md_dict3 = {'a': 2, 'c': 3}
    tt_base = Frame(np.ones((5, 3)), frame_no=frame_no, metadata=md_dict)

    tt = Frame(tt_base, frame_no=frame_no, metadata=md_dict2)
    target_dict = dict(md_dict)
    target_dict.update(md_dict2)
    # print(target_dict)
    # print(tt.metadata)
    assert tt.metadata == target_dict

    tt2 = Frame(tt_base, frame_no=frame_no, metadata=md_dict3)
    assert tt2.metadata == md_dict3



class FileLocker(object):
    """
    A context manager to lock and unlock a file

    See http://docs.python.org/2/library/contextlib.html
    http://docs.python.org/2/library/stdtypes.html#typecontextmanager
    http://docs.python.org/2/reference/datamodel.html#context-managers
    """
    def __init__(self, lock):
        self.lock = lock

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()
        return False


import os
import random
import types
import unittest
import pickle
import numpy as np
from numpy.testing import (assert_equal, assert_allclose)
import pims
import pytest

path, _ = os.path.split(os.path.abspath(__file__))
path = os.path.join(path, 'data')


def _skip_if_no_PyAV():
    import pims.pyav_reader
    if not pims.pyav_reader.available():
        raise unittest.SkipTest('PyAV not found. Skipping.')


def _skip_if_no_MoviePy():
    import pims.moviepy_reader
    if not pims.moviepy_reader.available():
        raise unittest.SkipTest('MoviePy not found. Skipping.')


def _skip_if_no_ImageIO_ffmpeg():
    import pims.imageio_reader
    if not pims.imageio_reader.ffmpeg_available():
        raise unittest.SkipTest('ImageIO and ffmpeg not found. Skipping.')


def _skip_if_no_libtiff():
    import pims.tiff_stack
    if not pims.tiff_stack.libtiff_available():
        raise unittest.SkipTest('libtiff not installed. Skipping.')


def _skip_if_no_tifffile():
    import pims.tiff_stack
    if not pims.tiff_stack.tifffile_available():
        raise unittest.SkipTest('tifffile not installed. Skipping.')


def _skip_if_no_skimage():
    try:
        import skimage
    except ImportError:
        raise unittest.SkipTest('skimage not installed. Skipping.')


def assert_image_equal(actual, expected):
    if np.issubdtype(actual.dtype, np.integer):
        assert_equal(actual, expected)
    else:
        if np.issubdtype(expected.dtype, np.integer):
            expected = expected/float(np.iinfo(expected.dtype).max)
        assert_allclose(actual, expected, atol=1/256.)


def save_dummy_png(filepath, filenames, shape):
    from PIL import Image
    if not os.path.isdir(filepath):
        os.mkdir(filepath)
    frames = []
    for f in filenames:
        dummy = np.random.randint(0, 255, shape).astype('uint8')
        im = Image.fromarray(dummy)
        im.save(os.path.join(filepath, f), 'png')
        frames.append(dummy)
        im.close()
    return frames


def clean_dummy_png(filepath, filenames):
    for f in filenames:
        os.remove(os.path.join(filepath, f))
    if os.listdir(filepath) == []:
        os.rmdir(filepath)


class _image_single(object):
    def check_skip(self):
        pass

    def test_bool(self):
        self.check_skip()
        pass

    def test_integer_attributes(self):
        self.check_skip()
        assert_equal(len(self.v.frame_shape), len(self.expected_shape))
        self.assertTrue(isinstance(self.v.frame_shape[0], int))
        self.assertTrue(isinstance(self.v.frame_shape[1], int))
        self.assertTrue(isinstance(len(self.v), int))

    def test_shape(self):
        self.check_skip()
        assert_equal(self.v.frame_shape, self.expected_shape)
        assert_equal(self.v[0].shape, self.expected_shape)

    def test_count(self):
        self.check_skip()
        assert_equal(len(self.v), self.expected_len)

    def test_repr(self):
        self.check_skip()
        # simple smoke test, values not checked
        repr(self.v)

    def tearDown(self):
        if hasattr(self, 'v'):
            self.v.close()


def box(letter):
    return pims.Frame(np.array(letter))


def assert_letters_equal(actual, expected):
    for actual_, expected_ in zip(actual, expected):
        # This contrived reader has weird shape behavior,
        # but that's not what I'm testing here.
        assert_equal(actual_.reshape((1, 1)), box(expected_).reshape((1, 1)))

def compare_slice_to_list(actual, expected):
    assert_letters_equal(actual, expected)
    # test lengths
    actual_len = len(actual)
    assert_equal(actual_len, len(expected))
    indices = list(range(len(actual)))
    for i in indices:
        # test positive indexing
        assert_letters_equal(actual[i], expected[i])
        # test negative indexing
        assert_letters_equal(actual[-i + 1], expected[-i + 1])
    # in reverse order
    for i in indices[::-1]:
        assert_letters_equal(actual[i], expected[i])
        assert_letters_equal(actual[-i + 1], expected[-i + 1])
    # in shuffled order (using a consistent random seed)
    r = random.Random(5)
    r.shuffle(indices)
    for i in indices:
        assert_letters_equal(actual[i], expected[i])
        assert_letters_equal(actual[-i + 1], expected[-i + 1])
    # test list indexing
    some_indices = [r.choice(indices) for _ in range(2)]
    assert_letters_equal(actual[some_indices],
                         np.array(expected)[some_indices])
    # mixing positive and negative indices
    some_indices = [r.choice(indices + [-i-1 for i in indices])
                    for _ in range(2)]
    assert_letters_equal(actual[some_indices],
                         np.array(expected)[some_indices])
    # test slices
    assert_letters_equal(actual[::2], expected[::2])
    assert_letters_equal(actual[1::2], expected[1::2])
    assert_letters_equal(actual[::3], expected[::3])
    assert_letters_equal(actual[1:], expected[1:])
    assert_letters_equal(actual[:], expected[:])
    assert_letters_equal(actual[:-1], expected[:-1])


class TestRecursiveSlicing(unittest.TestCase):

    def setUp(self):
        class DemoReader(pims.ImageSequence):
            def imread(self, filename, **kwargs):
                return np.array([[filename]])

        self.v = DemoReader(list('abcdefghij'))

    def test_slice_of_slice(self):
        slice1 = self.v[4:]
        compare_slice_to_list(slice1, list('efghij'))
        slice2 = slice1[-3:]
        compare_slice_to_list(slice2, list('hij'))
        slice1a = self.v[[3, 4, 5, 6, 7, 8, 9]]
        compare_slice_to_list(slice1a, list('defghij'))
        slice2a = slice1a[::2]
        compare_slice_to_list(slice2a, list('dfhj'))
        slice2b = slice1a[::-1]
        compare_slice_to_list(slice2b, list('jihgfed'))
        slice2c = slice1a[::-2]
        compare_slice_to_list(slice2c, list('jhfd'))
        # print('slice2d')
        slice2d = slice1a[:0:-1]
        compare_slice_to_list(slice2d, list('jihgfe'))
        slice2e = slice1a[-1:1:-1]
        compare_slice_to_list(slice2e, list('jihgf'))
        slice2f = slice1a[-2:1:-1]
        compare_slice_to_list(slice2f, list('ihgf'))
        slice2g = slice1a[::-3]
        compare_slice_to_list(slice2g, list('jgd'))
        slice2h = slice1a[[5, 6, 2, -1, 3, 3, 3, 0]]
        compare_slice_to_list(slice2h, list('ijfjgggd'))


    def test_slice_of_slice_of_slice(self):
        slice1 = self.v[4:]
        compare_slice_to_list(slice1, list('efghij'))
        slice2 = slice1[1:-1]
        compare_slice_to_list(slice2, list('fghi'))
        slice2a = slice1[[2, 3, 4]]
        compare_slice_to_list(slice2a, list('ghi'))
        slice3 = slice2[1::2]
        compare_slice_to_list(slice3, list('gi'))

    def test_slice_of_slice_of_slice_of_slice(self):
        # Take the red pill. It's slices all the way down!
        slice1 = self.v[4:]
        compare_slice_to_list(slice1, list('efghij'))
        slice2 = slice1[1:-1]
        compare_slice_to_list(slice2, list('fghi'))
        slice3 = slice2[1:]
        compare_slice_to_list(slice3, list('ghi'))
        slice4 = slice3[1:]
        compare_slice_to_list(slice4, list('hi'))

        # Give me another!
        slice1 = self.v[2:]
        compare_slice_to_list(slice1, list('cdefghij'))
        slice2 = slice1[0::2]
        compare_slice_to_list(slice2, list('cegi'))
        slice3 = slice2[:]
        compare_slice_to_list(slice3, list('cegi'))
        # print('define slice4')
        slice4 = slice3[:-1]
        # print('compare slice4')
        compare_slice_to_list(slice4, list('ceg'))
        # print('define slice4a')
        slice4a = slice3[::-1]
        # print('compare slice4a')
        compare_slice_to_list(slice4a, list('igec'))

    def test_slice_with_generator(self):
        slice1 = self.v[1:]
        compare_slice_to_list(slice1, list('bcdefghij'))
        slice2 = slice1[(i for i in range(2,5))]
        assert_letters_equal(slice2, list('def'))
        self.assertTrue(isinstance(slice2, types.GeneratorType))

def _rescale(img):
    return (img - img.min()) / img.ptp()

def _color_channel(img, channel):
    if img.ndim == 3:
        return img[:, :, channel]
    else:
        return img


class _image_series(_image_single):
    def test_iterator(self):
        self.check_skip()
        i = iter(self.v)
        assert_image_equal(next(i), self.frame0)
        assert_image_equal(next(i), self.frame1)

    def test_getting_slice(self):
        self.check_skip()
        tmp = list(self.v[0:2])
        frame0, frame1 = tmp
        assert_image_equal(frame0, self.frame0)
        assert_image_equal(frame1, self.frame1)

    def test_slice_of_slice(self):
        # More thorough recursive slicing tests, making use of more than
        # the two frames available for these tests, are elsewhere:
        # see test_recursive_slicing.
        self.check_skip()
        tmp = self.v[0:2]
        tmp1 = tmp[1:]
        frame1 = tmp1[0]
        assert_image_equal(frame1, self.frame1)

        # Do the same thing again, show that the generators are not dead.
        tmp1 = tmp[1:]
        frame1 = tmp1[0]
        assert_image_equal(frame1, self.frame1)

        frame0 = tmp[0]
        assert_image_equal(frame0, self.frame0)

        # Show that we can listify the slice twice.
        frame0, frame1 = list(tmp)
        assert_image_equal(frame0, self.frame0)
        assert_image_equal(frame1, self.frame1)
        frame0, frame1 = list(tmp)
        assert_image_equal(frame0, self.frame0)
        assert_image_equal(frame1, self.frame1)

    def test_pipeline_simple(self):
        rescale = pims.pipeline(_rescale)
        rescaled_v = rescale(self.v[:1])

        assert_image_equal(rescaled_v[0], _rescale(self.frame0))

    def test_pipeline_with_args(self):
        color_channel = pims.pipeline(_color_channel)
        red = color_channel(self.v, 0)
        green = color_channel(self.v, 1)

        assert_image_equal(red[0], _color_channel(self.frame0, 0))
        assert_image_equal(green[0], _color_channel(self.frame0, 1))

        # Multiple pipelines backed by the same data are indep,
        # so this call to red is unaffected by green above.
        assert_image_equal(red[0], _color_channel(self.frame0, 0))

    def test_composed_pipelines(self):
        color_channel = pims.pipeline(_color_channel)
        rescale = pims.pipeline(_rescale)

        composed = rescale(color_channel(self.v, 0))

        expected = _rescale(_color_channel(self.v[0], 0))
        assert_image_equal(composed[0], expected)

    def test_as_grey(self):
        gr = pims.as_grey(self.v)
        assert len(gr[0].shape) == 2
        # Calling a second time does nothing
        gr2 = pims.as_grey(gr)
        assert_image_equal(gr[0], gr2[0])

        # Alternate spelling accepted
        gr = pims.as_gray(self.v)
        assert len(gr[0].shape) == 2

    def test_getting_single_frame(self):
        self.check_skip()
        assert_image_equal(self.v[0], self.frame0)
        assert_image_equal(self.v[0], self.frame0)
        assert_image_equal(self.v[1], self.frame1)
        assert_image_equal(self.v[1], self.frame1)

    def test_getting_list(self):
        self.check_skip()
        actual = list(self.v[[1, 0, 0, 1, 1]])
        expected = [self.frame1, self.frame0, self.frame0, self.frame1,
                    self.frame1]
        [assert_image_equal(a, b) for a, b in zip(actual, expected)]

    def test_frame_number_present(self):
        self.check_skip()
        for frame_no in [0, 1, 2, 1]:
            self.assertTrue(hasattr(self.v[frame_no], 'frame_no'))
            not_none = self.v[frame_no].frame_no is not None
            self.assertTrue(not_none)

    def test_frame_number_accurate(self):
        self.check_skip()
        for frame_no in [0, 1, 2, 1]:
            self.assertEqual(self.v[frame_no].frame_no, frame_no)

    def test_simple_negative_index(self):
        self.check_skip()
        self.v[-1]
        list(self.v[[0, -1]])


class TestImageReaderTIFF(_image_single, unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(path, 'stuck.tif')
        self.frame0 = np.load(os.path.join(path, 'stuck_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'stuck_frame1.npy'))
        self.klass = pims.ImageReader
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (5, 512, 512)
        self.expected_len = 1


class TestImageReaderPNG(_image_single, unittest.TestCase):
    def setUp(self):
        self.klass = pims.ImageReader
        self.kwargs = dict()
        self.expected_shape = (10, 11)
        self.expected_len = 1

        save_dummy_png(path, ['dummy.png'], self.expected_shape)
        self.v = self.klass(os.path.join(path, 'dummy.png'), **self.kwargs)
        clean_dummy_png(path, ['dummy.png'])


class TestImageReaderND(_image_single, unittest.TestCase):
    def setUp(self):
        self.klass = pims.ImageReaderND
        self.kwargs = dict()
        self.expected_shape = (10, 11, 3)
        self.expected_len = 1

        save_dummy_png(path, ['dummy.png'], self.expected_shape)
        self.v = self.klass(os.path.join(path, 'dummy.png'), **self.kwargs)
        clean_dummy_png(path, ['dummy.png'])


class TestVideo_PyAV_timed(_image_series, unittest.TestCase):
    def check_skip(self):
        _skip_if_no_PyAV()

    def setUp(self):
        _skip_if_no_PyAV()
        self.filename = os.path.join(path, 'bulk-water.mov')
        self.frame0 = np.load(os.path.join(path, 'bulk-water_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'bulk-water_frame1.npy'))
        self.klass = pims.PyAVReaderTimed
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (424, 640, 3)
        self.expected_len = 480


class TestVideo_PyAV_indexed(_image_series, unittest.TestCase):
    def check_skip(self):
        _skip_if_no_PyAV()

    def setUp(self):
        _skip_if_no_PyAV()
        self.filename = os.path.join(path, 'bulk-water.mov')
        self.frame0 = np.load(os.path.join(path, 'bulk-water_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'bulk-water_frame1.npy'))
        self.klass = pims.PyAVReaderIndexed
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (424, 640, 3)
        self.expected_len = 480


class TestVideo_ImageIO(_image_series, unittest.TestCase):
    def check_skip(self):
        _skip_if_no_ImageIO_ffmpeg()

    def setUp(self):
        _skip_if_no_ImageIO_ffmpeg()
        self.filename = os.path.join(path, 'bulk-water.mov')
        self.frame0 = np.load(os.path.join(path, 'bulk-water_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'bulk-water_frame1.npy'))
        self.klass = pims.ImageIOReader
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (424, 640, 3)
        self.expected_len = 480

class TestVideo_MoviePy(_image_series, unittest.TestCase):
    def check_skip(self):
        _skip_if_no_MoviePy()

    def setUp(self):
        _skip_if_no_MoviePy()
        self.filename = os.path.join(path, 'bulk-water.mov')
        self.frame0 = np.load(os.path.join(path, 'bulk-water_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'bulk-water_frame1.npy'))
        self.klass = pims.MoviePyReader
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (424, 640, 3)
        self.expected_len = 480


class _tiff_image_series(_image_series):
    def test_metadata(self):
        m = self.v[0].metadata
        pkl_path = os.path.join(path, 'stuck_metadata_py3.pkl')
        with open(pkl_path, 'rb') as p:
            d = pickle.load(p)
        assert_equal(m, d)


class TestTiffStack_libtiff(_tiff_image_series, unittest.TestCase):
    def check_skip(self):
        _skip_if_no_libtiff()

    def setUp(self):
        _skip_if_no_libtiff()
        self.filename = os.path.join(path, 'stuck.tif')
        self.frame0 = np.load(os.path.join(path, 'stuck_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'stuck_frame1.npy'))
        self.klass = pims.TiffStack_libtiff
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (512, 512)
        self.expected_len = 5


class TestTiffStack_pil(_tiff_image_series, unittest.TestCase):
    def check_skip(self):
        pass

    def setUp(self):
        self.filename = os.path.join(path, 'stuck.tif')
        self.frame0 = np.load(os.path.join(path, 'stuck_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'stuck_frame1.npy'))
        self.klass = pims.TiffStack_pil
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (512, 512)
        self.expected_len = 5


class TestTiffStack_tifffile(_tiff_image_series, unittest.TestCase):
    def check_skip(self):
        pass

    def setUp(self):
        _skip_if_no_tifffile()
        self.filename = os.path.join(path, 'stuck.tif')
        self.frame0 = np.load(os.path.join(path, 'stuck_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'stuck_frame1.npy'))
        self.klass = pims.TiffStack_tifffile
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (512, 512)
        self.expected_len = 5


class TestSpeStack(_image_series, unittest.TestCase):
    def check_skip(self):
        pass

    def setUp(self):
        self.filename = os.path.join(path, 'spestack_test.spe')
        self.frame0 = np.load(os.path.join(path, 'spestack_test_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'spestack_test_frame1.npy'))
        self.klass = pims.SpeStack
        self.kwargs = dict()
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = (128, 128)
        self.expected_len = 5

    def test_metadata(self):
        m = self.v.metadata
        with open(os.path.join(path, 'spestack_test_metadata.pkl'), 'rb') as p:
            d = pickle.load(p, encoding="latin1")
            #spare4 is actually a byte array
            d["spare4"] = d["spare4"].encode("latin1")

        assert_equal(m, d)



class TestOpenFiles(unittest.TestCase):
    def test_open_png(self):
        self.filenames = ['dummy_png.png']
        shape = (10, 11)
        save_dummy_png(path, self.filenames, shape)
        pims.open(os.path.join(path, 'dummy_png.png')).close()
        clean_dummy_png(path, self.filenames)

    def test_open_pngs(self):
        self.filepath = os.path.join(path, 'image_sequence')
        self.filenames = ['T76S3F00001.png', 'T76S3F00002.png',
                          'T76S3F00003.png', 'T76S3F00004.png',
                          'T76S3F00005.png']
        shape = (10, 11)
        save_dummy_png(self.filepath, self.filenames, shape)
        pims.open(os.path.join(path, 'image_sequence', '*.png')).close()
        clean_dummy_png(self.filepath, self.filenames)

    def test_open_mov(self):
        _skip_if_no_PyAV()
        pims.open(os.path.join(path, 'bulk-water.mov')).close()

    def test_open_tiff(self):
        _skip_if_no_tifffile()
        pims.open(os.path.join(path, 'stuck.tif')).close()

if __name__ == '__main__':
    unittest.main()
